/*
 * Copyright 2014 Kai Uwe Broulik <kde@privat.broulik.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "androidmessagebox.h"

#include <QDebug>

#include <QtAndroid>
#include <QAndroidJniEnvironment>
#include <jni.h> // for the low level JNI stuff

AndroidMessageBox::AndroidMessageBox(QObject *parent)
    : QObject(parent)
    , m_rejectButton(QStringLiteral("Close")) // Don't do that ;)
{

}

AndroidMessageBox::~AndroidMessageBox()
{

}

void AndroidMessageBox::show()
{
    if (m_title.isEmpty() || m_text.isEmpty()) {
        return;
    }

    // This gives us a pointer to the JNI Environment, needed for eg. exception handling
    QAndroidJniEnvironment env;

    // Lazy init
    if (!m_javaMessageBox.isValid()) {
        // We instantiate our AndroidMessageBox java class
        // it can be found in the android/src/AndroidMessageBox.java file
        // 1. Fully qualified class name, ie. including package name, separated by slashes
        // 2. The signature of the constructors, they always return "void" to the JNI
        // 3. and 4. the actual parameters
        m_javaMessageBox = QAndroidJniObject("org/kde/akademy/android/AndroidMessageBox",
                                             "(Landroid/app/Activity;J)V",
                                             QtAndroid::androidActivity().object(),
                                             jlong(static_cast<QObject *>(this)));

        // Since this is Java it can throw Exceptions all over the place, and doing a subsequent JNI call
        // with an exception pending crashes your app, so you should really do that after each single JNI call :/
        // and I think it will even blow up if you do ExceptionClear() without checking if there is one first
        if (env->ExceptionCheck()) {
            qWarning() << "Exception instantiating MessageBox backing Java class";
            env->ExceptionClear();
            return;
        }
    }

    // Since JNI is relatively slow, try to reduce the amount of calls you're doing.
    // You could have getters and setters and stuff but I'd prefer sending everything over
    // at once and then be done with it
    m_javaMessageBox.callMethod<void>("show",
                                      "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V",
                                      QAndroidJniObject::fromString(m_title).object<jstring>(),
                                      QAndroidJniObject::fromString(m_text).object<jstring>(),
                                      QAndroidJniObject::fromString(m_rejectButton).object<jstring>(),
                                      QAndroidJniObject::fromString(m_acceptButton).object<jstring>());

    // And once again.. ;)
    if (env->ExceptionCheck()) {
        qWarning() << "Exception showing message box";
        env->ExceptionClear();
    }
}

// This holds the id of our class which you need to free it some time later
static jclass s_messageBoxClassId = 0;

// This is called by the JNI when you call the native method
// env and thiz are always there (and usually unused), and the third
// is our own parameter, you can have as many as you like obviously
static void messageBoxAccepted(JNIEnv *env, jobject thiz, jlong handler)
{
    Q_UNUSED(env);
    Q_UNUSED(thiz);

    // Do some nasty magic to get our instance
    QObject *object = reinterpret_cast<QObject *>(handler);
    // and then emit the signal directly
    QMetaObject::invokeMethod(object, "accepted", Qt::QueuedConnection);
}

static void messageBoxRejected(JNIEnv *env, jobject thiz, jlong handler)
{
    Q_UNUSED(env);
    Q_UNUSED(thiz);

    QObject *object = reinterpret_cast<QObject *>(handler);
    QMetaObject::invokeMethod(object, "rejected", Qt::QueuedConnection);
}

// The native methods, name of the method, its JNI-style signature
// note that this is the signature of the method above, J is long, and the first two parameters don't count
static JNINativeMethod s_messageBoxMethods[] = {
    {"accepted", "(J)V", (void *)messageBoxAccepted},
    {"rejected", "(J)V", (void *)messageBoxRejected}
};

// This thing actually registers all the native methods, it is called by the
// JavaVM on application startup
//
// In http://qt-project.org/doc/qt-5/qandroidjniobject.html#java-native-methods
// they call RegisterNatives() at an arbitrary location in the program,
// so please try if that also works and tell me :)
JNIEXPORT jint JNI_OnLoad(JavaVM *vm, void *reserved)
{
    Q_UNUSED(reserved);

    // Don't use QAndroidJniEnv just yet otherwise your qApp will be created in another thread
    // QAndroidJniEnvironment env;
    JNIEnv *env;
    if (vm->GetEnv(reinterpret_cast<void**>(&env), JNI_VERSION_1_6) != JNI_OK) {
        qCritical("Failed to get JNI Env");
        return JNI_ERR;
    }

    // First we need to find our class
    jclass messageBoxClass = env->FindClass("org/kde/akademy/android/AndroidMessageBox");
    if (!messageBoxClass) {
        qCritical("Could not find Android MessageBox Java class for registering natives");
        // returning somethng else than JNI_VERSION_FOO will cause the startup to abort
        return JNI_ERR;
    }

    // Create a global ref to our class. Be careful with this.
    // Use env->DeleteGlobalRef(s_messageBoxClassid) to free it
    s_messageBoxClassId = static_cast<jclass>(env->NewGlobalRef(messageBoxClass));

    // And finally register the native methods we declared above on the class we just "found"
    if (env->RegisterNatives(s_messageBoxClassId, s_messageBoxMethods, sizeof(s_messageBoxMethods) / sizeof(s_messageBoxMethods[0])) < 0) {
        qCritical("Failed to register natives for MessageBox");
        return JNI_ERR;
    }

    return JNI_VERSION_1_6;
}
