/*
 * Copyright 2014 Kai Uwe Broulik <kde@privat.broulik.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "androidshareprovider.h"

#include <QString>

#include <QtAndroid>
#include <QAndroidActivityResultReceiver>

// QAndroidActivityResultReceiver is an abstract class for receiving "results or activities",
// ie. the "return value" when the app resumes
class AndroidShareResultReceiver : public QAndroidActivityResultReceiver
{
public:
    // receiverRequestCode - a custom int you set (42 in our example)
    // resultCode - whether the request was successful or failed
    // data - an instance of android/content/Intent with the data
    void handleActivityResult(int receiverRequestCode, int resultCode, const QAndroidJniObject &data);

private:
    friend class AndroidShareProvider;
    // So we can emit signals directly on our share provider
    AndroidShareProvider *provider = nullptr;

};

void AndroidShareResultReceiver::handleActivityResult(int receiverRequestCode, int resultCode, const QAndroidJniObject &data)
{
    if (!provider) {
        return;
    }

    // We send 42 along and if we get something else, we're not interested
    // this is handy for having one receiver for different tasks
    if (receiverRequestCode != 42) {
        return;
    }

    QString actionName;

    // If QAndroidJniObject::isValid() return false, it means the backing Java object
    // is null and every call on it will yield a NullPointerException blowing up your app
    if (data.isValid()) {
        // Android documentation says "getAction" returns the type of the action, such as "Action.SEND",
        // however it actually returns the app that handled the request for some reason

        // Qt automatically determins the signature "()Ljava/lang/String;" based on the template type
        QAndroidJniObject action = data.callObjectMethod<jstring>("getAction");
        actionName = action.toString();
    }

    // Get all the constants so we don't hardcode them
    int resultCanceled = QAndroidJniObject::getStaticField<jint>("android/app/Activity", "RESULT_CANCELED");
    int resultOk = QAndroidJniObject::getStaticField<jint>("android/app/Activity", "RESULT_OK");

    // And finally emit the respective signals
    if (resultCode == resultCanceled) {
        emit provider->error(actionName);
    } else if (resultCode == resultOk) {
        emit provider->shared(actionName);
    }
}

AndroidShareProvider::AndroidShareProvider(QObject *parent)
    : QObject(parent)
    , m_receiver(new AndroidShareResultReceiver)
{
    m_receiver->provider = this;
}

AndroidShareProvider::~AndroidShareProvider()
{

}

void AndroidShareProvider::share(const QString &prompt, const QString &title, const QString &text)
{
    /*
    That is the Java code that we want to have:

    Intent intent = new Intent(Intent.ACTION_SEND);
    intent.setType("text/plain");
    intent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_WHEN_TASK_RESET);

    intent.putExtra(Intent.EXTRA_SUBJECT, title);
    intent.putExtra(Intent.EXTRA_TEXT, text);

    qtActivity.startActivity(Intent.createChooser(intent, prompt));
    */

    // Again, get all the constants first
    // get(Static)Field means it returns an integral type such as int, bool, etc
    // get(Static)ObjectField means it returns a QAndroidJniObject, note that String is also an object
    QAndroidJniObject actionSend = QAndroidJniObject::getStaticObjectField<jstring>("android/content/Intent", "ACTION_SEND");
    int flagActivityClearWhenTaskReset = QAndroidJniObject::getStaticField<jint>("android/content/Intent", "FLAG_ACTIVITY_CLEAR_WHEN_TASK_RESET");
    QAndroidJniObject extraSubject = QAndroidJniObject::getStaticObjectField<jstring>("android/content/Intent", "EXTRA_SUBJECT");
    QAndroidJniObject extraText = QAndroidJniObject::getStaticObjectField<jstring>("android/content/Intent", "EXTRA_TEXT");

    // Instantiate our Intent, constructors always return "void" to the JNI
    QAndroidJniObject intent("android/content/Intent", "(Ljava/lang/String;)V", actionSend.object());
    // Unfortunately for us, you can chain all those calls, which means they all return an instance
    // of android/content/Intent so we cannot use Qt's templated calls
    intent.callObjectMethod("setType", "(Ljava/lang/String;)Landroid/content/Intent;", QAndroidJniObject::fromString(QStringLiteral("text/plain")).object());
    intent.callObjectMethod("addFlags", "(I)Landroid/content/Intent;", flagActivityClearWhenTaskReset);

    if (!title.isEmpty()) {
        intent.callObjectMethod("putExtra", "(Ljava/lang/String;Ljava/lang/String;)Landroid/content/Intent;", extraSubject.object(), QAndroidJniObject::fromString(title).object());
    }
    intent.callObjectMethod("putExtra", "(Ljava/lang/String;Ljava/lang/String;)Landroid/content/Intent;", extraText.object(), QAndroidJniObject::fromString(text).object());

    // Then we call the static Intent.createChooser and give it the intent object we just created and our "prompt" text
    QAndroidJniObject chooser = QAndroidJniObject::callStaticObjectMethod("android/content/Intent",
                                                                          "createChooser",
                                                                          "(Landroid/content/Intent;Ljava/lang/CharSequence;)Landroid/content/Intent;",
                                                                          intent.object(),
                                                                          QAndroidJniObject::fromString(prompt).object());
    // Start the activitiy
    // the 42 (our receiverRequestCode) is mandatory for whatever reason
    // If the third parameter is null (default), it will just launch the activity without bothering about a response
    QtAndroid::startActivity(chooser, 42, m_receiver.data());
}
